from collections.abc import Mapping
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from . import bindings
from .engine_interface import Path
from .engine_interface import RawExecInfoType
from .engine_interface import TaskGraph
from .engine_interface import WorkflowEngineWithSerialization
from .node import NodeIdType
from .task import Task


class CoreWorkflowEngine(WorkflowEngineWithSerialization):

    def execute_graph(
        self,
        graph: Any,
        *,
        inputs: Optional[List[dict]] = None,
        load_options: Optional[dict] = None,
        varinfo: Optional[dict] = None,
        execinfo: RawExecInfoType = None,
        task_options: Optional[dict] = None,
        outputs: Optional[List[dict]] = None,
        merge_outputs: Optional[bool] = True,
        # Engine specific:
        output_tasks: Optional[bool] = False,
        raise_on_error: Optional[bool] = True,
    ) -> Union[Dict[NodeIdType, Task], Dict[str, Any]]:
        return bindings.execute_graph(
            graph,
            inputs=inputs,
            load_options=load_options,
            varinfo=varinfo,
            execinfo=execinfo,
            task_options=task_options,
            raise_on_error=raise_on_error,
            outputs=outputs,
            merge_outputs=merge_outputs,
            output_tasks=output_tasks,
        )

    def deserialize_graph(
        self,
        graph: Any,
        *,
        inputs: Optional[List[dict]] = None,
        representation: Optional[str] = None,
        root_dir: Optional[Union[str, Path]] = None,
        root_module: Optional[str] = None,
        # Serializer specific:
        **load_options,
    ) -> TaskGraph:
        return bindings.load_graph(
            graph,
            inputs=inputs,
            representation=representation,
            root_dir=root_dir,
            root_module=root_module,
            **load_options,
        )

    def serialize_graph(
        self,
        graph: TaskGraph,
        destination,
        *,
        representation: Optional[str] = None,
        # Serializer specific:
        **save_options,
    ) -> Union[str, dict]:
        return bindings.save_graph(
            graph, destination, representation=representation, **save_options
        )

    def get_graph_representation(self, graph: Any) -> Optional[str]:
        if isinstance(graph, Mapping):
            return "json_dict"
        if isinstance(graph, (str, Path)):
            if graph == "test_core":
                return "test_core"
            if isinstance(graph, str) and "{" in graph and "}" in graph:
                return "json_string"
            filename = str(graph).lower()
            if filename.endswith(".json"):
                return "json"
            elif filename.endswith((".yml", ".yaml")):
                return "yaml"



import logging
import os
import subprocess
import sys
from typing import Mapping

from .task import Task

SCRIPT_ARGUMENT = "_script"
WIN32 = sys.platform == "win32"


logger = logging.getLogger(__name__)


class ScriptExecutorTask(
    Task,
    input_names=[SCRIPT_ARGUMENT],
    optional_input_names=["_capture_output", "_merge_err", "_raise_on_error"],
    output_names=["return_code", "out", "err"],
):
    """Task wrapper for a shell command or script.

    When the :code:`SCRIPT_ARGUMENT` input variable is not an existing file, it is assumed to be a command.

    When it is a file, it is assumed to be executable when

    - Windows: it does not have the `".py"` file name extension
    - Linux/Mac: the content does not start with a shebang "#!"

    When the file is not executable a command is prepended

    - :code:`sys.executable` when the file name has the `".py"` extension
    - :code:`bash` otherwise

    Examples of different types of command line arguments

    - **Single-character argument names**:

        .. code-block:: bash

            python -c "print('hello')"

        .. code-block:: python

            inputs = {SCRIPT_ARGUMENT: "python", "c": "print('hello')"}

    - **Arguments without a value**:

        .. code-block:: bash

            ls -a

        .. code-block:: python

            inputs = {SCRIPT_ARGUMENT: "ls", "a": ""}

    - **Multi-character argument names**:

        .. code-block:: bash

            ls --all

        .. code-block:: python

            inputs = {SCRIPT_ARGUMENT: "ls", "all": ""}

    - **Arguments without a name**:

        .. code-block:: bash

            ls .

        .. code-block:: python

            inputs = {SCRIPT_ARGUMENT: "ls", 0: "."}

    - **Merged single-character argument names**:

        .. code-block:: bash

            ls -ltrh

        Since we need `-ltrh` instead of `--ltrh` we can specify
        it as a positional argument

        .. code-block:: python

            inputs = {SCRIPT_ARGUMENT: "ls", 0: "-ltrh"}
    """

    SCRIPT_ARGUMENT = SCRIPT_ARGUMENT

    def _get_task_identifier(self, inputs: Mapping) -> str:
        return inputs.get(self.SCRIPT_ARGUMENT, self.class_registry_name())

    def run(self):
        fullname = self.inputs._script
        if not isinstance(fullname, str):
            raise TypeError(fullname, type(fullname))

        # Is script?
        if os.path.isfile(fullname):
            # existing python or shell script
            is_python = fullname.endswith(".py")
            fullname = os.path.abspath(fullname)
            if WIN32:
                # Assume all non-python scripts are executable
                is_executable = not is_python
            else:
                # Scripts with a shebang are executable
                with open(fullname, "r") as f:
                    is_executable = f.readline().startswith("#!")
        else:
            # command (although it could be a script that does not exist)
            is_python = False
            is_executable = True
            fullname = fullname.split(" ")

        # Select executable when fullname itself is not executable
        executable = None
        if not is_executable:
            if is_python:
                executable = sys.executable
            elif not WIN32:
                executable = "bash"

        # Command starts with "[executable] fullname ..."
        cmd = []
        if executable:
            cmd.append(executable)
        if isinstance(fullname, str):
            cmd.append(fullname)
        else:
            cmd.extend(fullname)

        # Script/command arguments
        skip = self.input_names()
        positional = list()
        for k, v in self.get_input_values().items():
            if k in skip:
                continue
            value = str(v)
            if isinstance(k, int):
                positional.append((k, value))
                continue
            else:
                if len(k) == 1:
                    argmarker = "-"
                else:
                    argmarker = "--"
                if value:
                    cmd.extend((argmarker + k, value))
                else:
                    cmd.append(argmarker + k)
        cmd.extend([v for _, v in sorted(positional)])

        logger.debug("Command: '%s'", " ".join(cmd))

        # Run
        stdout = stderr = None
        if self.inputs._capture_output:
            stdout = subprocess.PIPE
            if self.inputs._merge_err:
                stderr = subprocess.STDOUT
            else:
                stderr = subprocess.PIPE

        result = subprocess.run(cmd, cwd=os.getcwd(), stdout=stdout, stderr=stderr)
        if self.inputs._raise_on_error:
            result.check_returncode()

        self.outputs.return_code = result.returncode
        if result.stdout:
            self.outputs.out = result.stdout.decode()
        elif self.inputs._capture_output:
            self.outputs.out = ""
        else:
            self.outputs.out = None
        if result.stderr:
            self.outputs.err = result.stderr.decode()
        elif self.inputs._capture_output and not self.inputs._merge_err:
            self.outputs.err = ""
        else:
            self.outputs.err = None



from ...model import BaseInputModel
from ...model import BaseOutputModel
from ...task import Task


class MyTask3(
    Task,
    input_names=["z", "c"],
    optional_input_names=["x", "d"],
    output_names=["result", "error"],
):
    """Test 3"""

    def run(self):
        pass


def run(z, c, x=None, d=None):
    """Test"""
    pass


def myfunc(z, c, x=None, d=None):
    pass


class Task4Inputs(BaseInputModel):
    a: int
    b: float
    c: int = 0
    d: str = "DEFAULT"


class MyTask4(Task, input_model=Task4Inputs):
    pass


class Task5Outputs(BaseOutputModel):
    a: int
    b: float
    c: int = 0
    d: str = "DEFAULT"


class MyTask5(Task, output_model=Task5Outputs):
    pass



from collections import Counter
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import networkx

from ... import events
from ...inittask import add_dynamic_inputs
from ...inittask import instantiate_task as _instantiate_task
from ...node import NodeIdType
from ...task import Task
from .. import analysis
from .. import graph_io


def instantiate_task(graph: networkx.DiGraph, node_id: NodeIdType, **kw) -> Task:
    """Named arguments are dynamic input and Variable config.
    Default input from the persistent representation are added internally.
    """
    # Dynamic input has priority over default input
    nodeattrs = graph.nodes[node_id]
    return _instantiate_task(node_id, nodeattrs, **kw)


def instantiate_task_static(
    graph: networkx.DiGraph,
    node_id: NodeIdType,
    tasks: Optional[Dict[Task, int]] = None,
    varinfo: Optional[dict] = None,
    execinfo: Optional[dict] = None,
    task_options: Optional[dict] = None,
    evict_result_counter: Optional[Dict[NodeIdType, int]] = None,
) -> Task:
    """Instantiate destination task while no access to the dynamic inputs.
    Side effect: `tasks` will contain all predecessors.
    """
    if analysis.graph_is_cyclic(graph):
        raise RuntimeError("cannot execute cyclic graphs with ewokscore")
    if tasks is None:
        tasks = dict()
    if evict_result_counter is None:
        evict_result_counter = dict()
    # Input from previous tasks (instantiate them if needed)
    dynamic_inputs = dict()
    for source_id in analysis.node_predecessors(graph, node_id):
        source_task = tasks.get(source_id, None)
        if source_task is None:
            source_task = instantiate_task_static(
                graph,
                source_id,
                tasks=tasks,
                varinfo=varinfo,
                execinfo=execinfo,
                task_options=task_options,
                evict_result_counter=evict_result_counter,
            )
        link_attrs = graph[source_id][node_id]
        add_dynamic_inputs(
            dynamic_inputs,
            link_attrs,
            source_task.output_variables,
            source_id=source_id,
            target_id=node_id,
        )
        # Evict intermediate results
        if evict_result_counter:
            evict_result_counter[source_id] -= 1
            if evict_result_counter[source_id] == 0:
                tasks.pop(source_id)
    # Instantiate the requested task
    target_task = instantiate_task(
        graph,
        node_id,
        inputs=dynamic_inputs,
        varinfo=varinfo,
        execinfo=execinfo,
        task_options=task_options,
    )
    tasks[node_id] = target_task
    return target_task


def successor_counter(graph: networkx.DiGraph) -> Dict[NodeIdType, int]:
    nsuccessor = Counter()
    for edge in graph.edges:
        nsuccessor[edge[0]] += 1
    return nsuccessor


def execute_graph(
    graph: networkx.DiGraph,
    varinfo: Optional[dict] = None,
    execinfo: Optional[dict] = None,
    task_options: Optional[dict] = None,
    raise_on_error: Optional[bool] = True,
    outputs: Optional[List[dict]] = None,
    merge_outputs: Optional[bool] = True,
    output_tasks: Optional[bool] = False,
) -> Union[Dict[NodeIdType, Task], Dict[str, Any]]:
    """Sequential execution of DAGs.

    When `output_tasks` is `True` the arguments `outputs` and `merge_outputs`
    are ignored and instead of returning output values, it returns `Task` instances.
    This was introduced for testing.
    """
    with events.workflow_context(execinfo, workflow=graph) as execinfo:
        if analysis.graph_is_cyclic(graph):
            raise RuntimeError("cannot execute cyclic graphs")
        if analysis.graph_has_conditional_links(graph):
            raise RuntimeError("cannot execute graphs with conditional links")

        # Pepare containers for local state
        tasks = dict()
        if output_tasks:
            output_values = None
            evict_result_counter = None
        else:
            outputs = graph_io.parse_outputs(graph, outputs)
            output_values = dict()
            evict_result_counter = successor_counter(graph)

        # Execute in topological order
        for node_id in analysis.topological_sort(graph):
            task = instantiate_task_static(
                graph,
                node_id,
                tasks=tasks,
                varinfo=varinfo,
                execinfo=execinfo,
                task_options=task_options,
                evict_result_counter=evict_result_counter,
            )
            task.execute(
                raise_on_error=raise_on_error,
                cleanup_references=evict_result_counter is not None,
            )
            if execinfo:
                execinfo.setdefault("exception", task.exception)
            if not output_tasks:
                graph_io.add_output_values(
                    output_values, node_id, task, outputs, merge_outputs=merge_outputs
                )

        # Return results or Task instances
        if output_tasks:
            return tasks
        else:
            return output_values



from ...task import Task


class MyTask1(
    Task, input_names=["a"], optional_input_names=["b"], output_names=["result"]
):
    """Test 1"""

    def run(self):
        pass


class MyTask2(
    Task, input_names=["a"], optional_input_names=["b"], output_names=["result"]
):
    def run(self):
        pass


class MyTask3(Task, n_required_positional_inputs=2):
    def run(self):
        pass


def run(a, b=None):
    """Test 2"""
    pass


def myfunc(a, b=None):
    pass


def func_with_pos(a, b, /, c):
    pass


def _myfunc(a, b=None):
    pass


class _MyTask2(Task):
    def run(self):
        pass


class UnregisteredTask(Task, register=False):
    def run(self):
        pass



import json
from pathlib import Path

import pytest
import yaml

from ..graph import load_graph


@pytest.mark.parametrize("with_ext", [True, False])
@pytest.mark.parametrize("with_representation", [True, False])
@pytest.mark.parametrize("path_format", [str, Path])
def test_graph_discovery_json(with_ext, with_representation, tmp_path, path_format):
    _dump_graph_and_subgraph(tmp_path, "json", with_ext)

    ewoksgraph = load_graph(
        source=path_format("graph"),
        representation="json" if with_representation else None,
        root_dir=path_format(tmp_path),
    )

    assert set(ewoksgraph.graph.nodes) == {"node1", ("node2", "subnode1")}


@pytest.mark.parametrize("with_ext", [True, False])
@pytest.mark.parametrize("with_representation", [True, False])
@pytest.mark.parametrize("path_format", [str, Path])
def test_graph_discovery_yaml(with_ext, with_representation, tmp_path, path_format):
    _dump_graph_and_subgraph(tmp_path, "yaml", with_ext)

    ewoksgraph = load_graph(
        source=path_format("graph"),
        representation="yaml" if with_representation else None,
        root_dir=path_format(tmp_path),
    )

    assert set(ewoksgraph.graph.nodes) == {"node1", ("node2", "subnode1")}


@pytest.mark.parametrize("with_representation", [True, False])
def test_graph_discovery_json_module(with_representation):
    if with_representation:
        source = "ewokscore.tests.examples.loadtest.graph"
        representation = "json_module"
    else:
        source = "graph"
        representation = None

    ewoksgraph = load_graph(
        source=source,
        representation=representation,
        root_module="ewokscore.tests.examples.loadtest",
    )

    assert set(ewoksgraph.graph.nodes) == {"node1", ("node2", "subnode1")}


def _dump_graph_and_subgraph(tmp_path, format, with_ext):
    if format == "yaml":
        dump = yaml.dump

    if format == "json":
        dump = json.dump

    ext = f".{format}" if with_ext else ""
    with open(tmp_path / f"subgraph{ext}", mode="w") as f:
        dump(_SUBGRAPH, f)
    with open(tmp_path / f"graph{ext}", mode="w") as f:
        dump(_GRAPH, f)


_SUBGRAPH = {
    "graph": {
        "id": "subgraph",
        "schema_version": "1.1",
        "input_nodes": [{"id": "in", "node": "subnode1"}],
    },
    "nodes": [
        {
            "id": "subnode1",
            "task_type": "method",
            "task_identifier": "dummy",
            "default_inputs": [
                {"name": "name", "value": "subnode1"},
                {"name": "value", "value": 0},
            ],
        }
    ],
}

_GRAPH = {
    "graph": {"id": "graph", "schema_version": "1.1"},
    "nodes": [
        {
            "id": "node1",
            "task_type": "method",
            "task_identifier": "dummy",
            "default_inputs": [
                {"name": "name", "value": "node1"},
                {"name": "value", "value": 0},
            ],
        },
        {"id": "node2", "task_type": "graph", "task_identifier": "subgraph"},
    ],
    "links": [
        {
            "source": "node1",
            "target": "node2",
            "sub_target": "in",
            "data_mapping": [
                {"target_input": "value", "source_output": "return_value"}
            ],
        }
    ],
}
